/*
 * Copyright (c) 2019-2021 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Fri Jun 14 2019
 */
#include "autoconf.h"

#include "tk3-paparazzi.h"

/* settings */
tk3setting(TK3SET_MOTORS, 6, "number of controlled motors", 10);


/* io buffers */
static uint8_t tk3_nocache(mem[2][128]);
static struct tk3iob inb = tk3iob_init("mkbl input", mem[0], 32);
static struct tk3iob outb = tk3iob_init("mkbl output", mem[1], 32);


/* I2C communication state */
static struct {
  uint32_t sladdr, motors;
  uint32_t active;
} rx;

static struct {
  const stm32_dma_stream_t *dma;
  uint32_t len;
  uint32_t nack;
} tx;

static tk3timer(restart);

static void	mkbl_restart(void *arg);
static void	mkbl_complete(size_t wlen);
static void	mkbl_start(void);


/* --- tk3mkbl_init -------------------------------------------------------- */

void
tk3mkbl_init()
{
  /* reset I2C */
  tk3mkbl_fini();
  rccResetI2C2();
  rccEnableI2C2(true);
  rx.motors = tk3set_get(TK3SET_MOTORS, 6);
  rx.sladdr = 1;
  rx.active = 0;

  /* bitbang 9 clock pulses at 100kHz to unlock any locked slaves */
  palSetPad(GPIOB, 10);
  palSetPad(GPIOB, 11);
  palSetPadMode(GPIOB, 10, PAL_MODE_OUTPUT_OPENDRAIN);
  palSetPadMode(GPIOB, 11, PAL_MODE_OUTPUT_OPENDRAIN);
  for (int i = 0; i < 9; i++) {
    tk3clk_delay(5);
    palClearPad(GPIOB, 10);
    tk3clk_delay(5);
    palSetPad(GPIOB, 10);
  }

  /* configure SDA/SCL lines */
  palSetPadMode(GPIOB, 10, PAL_MODE_ALTERNATE(4) | PAL_STM32_OTYPE_OPENDRAIN);
  palSetPadMode(GPIOB, 11, PAL_MODE_ALTERNATE(4) | PAL_STM32_OTYPE_OPENDRAIN);

  /* DMA */
  tx.dma = dmaStreamAlloc(STM32_I2C_I2C2_TX_DMA_STREAM, 0, NULL, NULL);
  dmaStreamSetMode(
    tx.dma,
    STM32_DMA_CR_PSIZE_BYTE | STM32_DMA_CR_MSIZE_BYTE |
    STM32_DMA_CR_MINC | STM32_DMA_CR_DIR_M2P |
    STM32_DMA_CR_CHSEL(
      STM32_DMA_GETCHANNEL(
        STM32_I2C_I2C2_TX_DMA_STREAM, STM32_I2C2_TX_DMA_CHN)) |
    STM32_DMA_CR_PL(STM32_I2C_I2C2_DMA_PRIORITY));
  dmaStreamSetPeripheral(tx.dma, &I2C2->TXDR);
  dmaStreamDisable(tx.dma);

  /* 500kHz bus */
  I2C2->TIMINGR =
    ( 0 << I2C_TIMINGR_PRESC_Pos) |
    (14 << I2C_TIMINGR_SCLDEL_Pos) |
    ( 2 << I2C_TIMINGR_SDADEL_Pos) |
    (16 << I2C_TIMINGR_SCLH_Pos) |
    (16 << I2C_TIMINGR_SCLL_Pos);

  /* GC address */
  I2C2->CR2 = 0;

  /* 310ms SCL low timeout */
  I2C2->TIMEOUTR =
    4095 | I2C_TIMEOUTR_TIMOUTEN;

  /* interrupts */
  nvicEnableVector(STM32_I2C2_EVENT_NUMBER, STM32_I2C_I2C2_IRQ_PRIORITY);
  nvicEnableVector(STM32_I2C2_ERROR_NUMBER, STM32_I2C_I2C2_IRQ_PRIORITY);
  I2C2->CR1 =
    I2C_CR1_TXDMAEN |
    I2C_CR1_TCIE | I2C_CR1_RXIE |
    I2C_CR1_STOPIE | I2C_CR1_ERRIE;

  /* enable peripheral */
  I2C2->CR1 |= I2C_CR1_PE;

  /* start */
  tk3msg_log(TK3CH_USB0, "Nmotor bus configured");
  mkbl_start();
}

static void
mkbl_restart(void *arg)
{
  tk3fb_off(TK3FB_MKBL_ERR);
  tk3mkbl_init();
}


/* --- tk3mkbl_fini -------------------------------------------------------- */

void
tk3mkbl_fini()
{
  tk3clk_deltimer(&restart);
  nvicDisableVector(STM32_I2C2_EVENT_NUMBER);
  nvicDisableVector(STM32_I2C2_ERROR_NUMBER);
  I2C2->CR1 &= ~I2C_CR1_PE;
  dmaStreamFree(tx.dma);

  rccDisableI2C2();
  palSetPadMode(GPIOB, 10, PAL_MODE_RESET);
  palSetPadMode(GPIOB, 11, PAL_MODE_RESET);
  tk3fb_off(TK3FB_MKBL_ERR);
}


/* --- tk3mkbl_iob --------------------------------------------------------- */

struct tk3iob *
tk3mkbl_iniob()
{
  return &inb;
}

struct tk3iob *
tk3mkbl_outiob()
{
  return &outb;
}


/* --- mkbl_complete ------------------------------------------------------- */

static void
mkbl_complete(size_t wlen)
{
  if (wlen) {
    dmaStreamDisable(tx.dma);
    tk3iob_rcommit(&outb, wlen);
  }

  if (tx.nack)
    tx.nack = 0;
  else if (wlen)
    tk3fb_flash(TK3FB_MKBL_0, 1000);
  else
    tk3fb_flash(TK3FB_MKBL_0 + rx.sladdr, 1000);

  mkbl_start();
}


/* --- mkbl_start ---------------------------------------------------------- */

static void
mkbl_start()
{
  struct tk3iob_extent e;

  e = tk3iob_rbegin(&outb, 32);
  tx.len = e.len;
  if (e.len) {
    /* master transmit, general call */
    dmaStreamSetMemory0(tx.dma, e.r);
    dmaStreamSetTransactionSize(tx.dma, e.len);
    dmaStreamEnable(tx.dma);

    rx.active = 0;
    I2C2->CR2 = I2C_CR2_START | (e.len << I2C_CR2_NBYTES_Pos);
  } else {
    /* master receive, current slave address - at most 0xff bytes can be
     * received, but there is no such big message */
    if (!--rx.sladdr) rx.sladdr = rx.motors;
    rx.active = 1;
    I2C2->CR2 =
      I2C_CR2_START |
      ((0x10 + rx.sladdr) << 1) | I2C_CR2_RD_WRN |
      (0xff << I2C_CR2_NBYTES_Pos);
  }
}


/* --- STM32_I2C2_EVENT_HANDLER -------------------------------------------- */

OSAL_IRQ_HANDLER(STM32_I2C2_EVENT_HANDLER)
{
  uint32_t isr = I2C2->ISR;
  I2C2->ICR = isr & (I2C_ICR_STOPCF | I2C_ICR_NACKCF);

  /* nack received: nop */
  if (isr & I2C_ISR_NACKF) tx.nack = 1;

  /* data received */
  if (isr & I2C_ISR_RXNE) {
    uint8_t c = I2C2->RXDR;

    if (rx.active) {
      tk3iob_putc(&inb, c);
      if (c == '$') {
        I2C2->CR2 |= I2C_CR2_STOP;
        rx.active = 0;
      }
      return;
    }
    /* discard extra meaningless byte (only one) after '$' */
    /* FALLTHROUGH in case stop condition was already detected */
  }

  /* transfer complete or stop condition */
  if (isr & (I2C_ISR_TC | I2C_ISR_STOPF)) {
    mkbl_complete(tx.len);
    return;
  }
}

/* --- STM32_I2C2_ERROR_HANDLER -------------------------------------------- */

OSAL_IRQ_HANDLER(STM32_I2C2_ERROR_HANDLER)
{
  uint32_t isr;

  isr = I2C2->ISR & (I2C_ISR_TIMEOUT | I2C_ISR_BERR | I2C_ISR_ARLO);
  I2C2->ICR = isr;

  /* reset bus state */
  tk3mkbl_fini();

  /* start over after 5s */
  tk3fb_on(TK3FB_MKBL_ERR);
  tk3clk_settimer(&restart, 0, 5000000, TK3EV_NOW, mkbl_restart, NULL);
  tk3msg_log(TK3CH_USB0, "Emotor bus error");
}
