/*
 * Copyright (c) 2019-2021 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Tue Jul 23 2019
 */
#include "autoconf.h"

#include <stdarg.h>
#include <string.h>
#include <strings.h>

#include "tk3-paparazzi.h"

/* active sensors */
static struct {
  int active;
  int mpu9250, ak8963, rm3100;
} sensorp;

/* reader jobs */
static struct {
  struct {
    uint32_t chans;
    uint8_t seq;
  } imu, mag;
} reader;

static tk3timer(restart);

static void	tk3sens_restart(void *arg);
static void	tk3sens_reader(void *iostatus);


/* --- tk3sens_init -------------------------------------------------------- */

void
tk3sens_init()
{
  int s = 0;
  tk3sens_fini();

  tk3fb_flash(TK3FB_SENS, 100000);
  tk3fb_on(TK3FB_SENS_BUSY);
  tk3sens_hwinit();

  /* probe sensors */
  if (tk3sens_mpu9250_probe()) {
    if (!tk3sens_mpu9250_init())
      sensorp.mpu9250 = 1;
  } else {
    tk3msg_log(TK3CH_USB0, "Eno IMU detected");
    s = 1;
  }

  if (tk3sens_rm3100_probe()) {
    if (!tk3sens_rm3100_init())
      sensorp.rm3100 = 1;
  } else if (tk3sens_ak8963_probe()) {
    if (!tk3sens_ak8963_init())
      sensorp.ak8963 = 1;
  } else {
    tk3msg_log(TK3CH_USB0, "Eno magnetometer detected");
    s = 1;
  }

  tk3fb_off(TK3FB_SENS_BUSY);
  if (s) {
    /* start over after 5s */
    tk3sens_restart(NULL);
  } else
    sensorp.active = 1;
}

static void
tk3sens_restart(void *arg)
{
  if (arg) {
    tk3sens_init();
    return;
  }

  tk3sens_fini();
  tk3fb_on(TK3FB_SENS_ERR);
  tk3clk_settimer(
    &restart, 0, 5000000, TK3EV_IDLE, tk3sens_restart, tk3sens_restart);
}


/* --- tk3sens_fini -------------------------------------------------------- */

void
tk3sens_fini()
{
  sensorp.active = 0;
  tk3clk_deltimer(&restart);
  memset(&sensorp, 0, sizeof(sensorp));
  tk3sens_hwfini();
  tk3fb_off(TK3FB_SENS_ERR);
  tk3fb_off(TK3FB_SENS_BUSY);
}


/* --- tk3sens_send* ------------------------------------------------------- */

void
tk3sens_sendimu(void *arg)
{
  enum tk3ch channel = (enum tk3ch)arg;

  if (!sensorp.active) return;

  reader.imu.chans |= 1<<channel;
  tk3sens_reader(NULL);
}

void
tk3sens_sendmag(void *arg)
{
  enum tk3ch channel = (enum tk3ch)arg;

  if (!sensorp.active) return;

  reader.mag.chans |= 1<<channel;
  tk3sens_reader(NULL);
}


/* --- tk3sens_reader ------------------------------------------------------ */

#define coroutine()                                                     \
  static uint32_t state_;                                               \
  for(uint32_t do_ = 1; do_; do_ = state_ = 0) switch(state_) default:

#define yield(x) \
  do { state_ = __LINE__; return x; case __LINE__:; } while (0)

static void
tk3sens_reader(void *iostatus)
{
  unsigned int chans;

  coroutine() {
    while (reader.imu.chans || reader.mag.chans) {
      /* IMU */
      if (reader.imu.chans) {
        int32_t a[3], w[3];

        reader.imu.seq++;
        if (sensorp.mpu9250) {
          tk3fb_on(TK3FB_SENS);
          tk3sens_mpu9250_aioread(tk3sens_reader);
          do yield(); while (!iostatus);
          tk3fb_off(TK3FB_SENS);
        }

        chans = reader.imu.chans;
        reader.imu.chans = 0;
        if (!iostatus || *(uint32_t *)iostatus) {
          tk3sens_restart(NULL);
          break;
        }

        if (!tk3sens_mpu9250_data(a, w)) {
          for(; chans; chans &= chans - 1 /* clear LSB */) {
            tk3msg_log(ffs(chans)-1, "I%1%2%2%2%2%2%2",
                       reader.imu.seq, a[0], a[1], a[2], w[0], w[1], w[2]);
          }
        }
      }

      /* magnetometer */
      if (reader.mag.chans) {
        int32_t m[3];

        reader.mag.seq++;
        if (sensorp.rm3100) {
          tk3fb_on(TK3FB_SENS);
          tk3sens_rm3100_aioread(tk3sens_reader);
          do yield(); while (!iostatus);
          tk3fb_off(TK3FB_SENS);
        } else if (sensorp.ak8963) {
          tk3fb_on(TK3FB_SENS);
          tk3sens_ak8963_aioread(tk3sens_reader);
          do yield(); while (!iostatus);
          tk3fb_off(TK3FB_SENS);
        }

        chans = reader.mag.chans;
        reader.mag.chans = 0;
        if (!iostatus || *(uint32_t *)iostatus) {
          tk3sens_restart(NULL);
          break;
        }

        if (sensorp.rm3100) {
          if (tk3sens_rm3100_data(m)) continue;
        } else if (sensorp.ak8963) {
          if (tk3sens_ak8963_data(m)) continue;
        }

        for(; chans; chans &= chans - 1 /* clear LSB */) {
          tk3msg_log(ffs(chans)-1, "C%1%2%2%2",
                     reader.mag.seq, m[0], m[1], m[2]);
        }
      }
    }
  }
}
